package ru.dbotthepony.kstarbound.server.world

import it.unimi.dsi.fastutil.io.FastByteArrayInputStream
import it.unimi.dsi.fastutil.io.FastByteArrayOutputStream
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kommons.arrays.Object2DArray
import ru.dbotthepony.kommons.io.BTreeDB6
import ru.dbotthepony.kommons.io.ByteKey
import ru.dbotthepony.kommons.io.readVarInt
import ru.dbotthepony.kommons.io.writeVarInt
import ru.dbotthepony.kommons.util.KOptional
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.tile.BuiltinMetaMaterials
import ru.dbotthepony.kstarbound.defs.tile.isNullTile
import ru.dbotthepony.kstarbound.io.BTreeDB5
import ru.dbotthepony.kstarbound.json.VersionedJson
import ru.dbotthepony.kstarbound.math.vector.Vector2i
import ru.dbotthepony.kstarbound.util.CarriedExecutor
import ru.dbotthepony.kstarbound.world.CHUNK_SIZE
import ru.dbotthepony.kstarbound.world.ChunkPos
import ru.dbotthepony.kstarbound.world.ChunkState
import ru.dbotthepony.kstarbound.world.WorldGeometry
import ru.dbotthepony.kstarbound.world.api.AbstractCell
import ru.dbotthepony.kstarbound.world.api.ImmutableCell
import ru.dbotthepony.kstarbound.world.api.MutableCell
import ru.dbotthepony.kstarbound.world.entities.AbstractEntity
import ru.dbotthepony.kstarbound.world.entities.tile.WorldObject
import java.io.BufferedInputStream
import java.io.BufferedOutputStream
import java.io.ByteArrayInputStream
import java.io.DataInputStream
import java.io.DataOutputStream
import java.io.File
import java.lang.ref.Cleaner
import java.lang.ref.WeakReference
import java.sql.Connection
import java.sql.DriverManager
import java.sql.PreparedStatement
import java.util.concurrent.CompletableFuture
import java.util.concurrent.CopyOnWriteArrayList
import java.util.concurrent.TimeUnit
import java.util.concurrent.locks.ReentrantLock
import java.util.function.Function
import java.util.function.Supplier
import java.util.zip.DeflaterOutputStream
import java.util.zip.InflaterInputStream
import kotlin.concurrent.withLock

sealed class LegacyWorldStorage() : WorldStorage() {
	protected abstract fun load(at: ByteKey): CompletableFuture<KOptional<ByteArray>>
	protected abstract fun write(at: ByteKey, value: ByteArray)

	override fun loadCells(pos: ChunkPos): CompletableFuture<KOptional<Pair<Object2DArray<out AbstractCell>, ChunkState>>> {
		val chunkX = pos.x
		val chunkY = pos.y
		val key = ByteKey(1, (chunkX shr 8).toByte(), chunkX.toByte(), (chunkY shr 8).toByte(), chunkY.toByte())

		return load(key).thenApplyAsync(Function {
			it.map {
				val reader = DataInputStream(BufferedInputStream(InflaterInputStream(ByteArrayInputStream(it))))
				val generationLevel = reader.readVarInt()
				val tileSerializationVersion = reader.readVarInt()

				val state = when (generationLevel) {
					0 -> ChunkState.EMPTY
					1 -> ChunkState.TERRAIN
					2 -> ChunkState.MICRO_DUNGEONS
					3 -> ChunkState.CAVE_LIQUID
					4 -> ChunkState.FULL
					else -> ChunkState.FULL
				}

				val result = Object2DArray.nulls<ImmutableCell>(CHUNK_SIZE, CHUNK_SIZE)

				for (y in 0 until CHUNK_SIZE) {
					for (x in 0 until CHUNK_SIZE) {
						val read = MutableCell().readLegacy(reader, tileSerializationVersion)

						if (read.foreground.material == BuiltinMetaMaterials.STRUCTURE) {
							read.foreground.material = BuiltinMetaMaterials.EMPTY
						}

						if (read.background.material == BuiltinMetaMaterials.STRUCTURE) {
							read.background.material = BuiltinMetaMaterials.EMPTY
						}

						result[x, y] = read.immutable()
					}
				}

				reader.close()
				result as Object2DArray<out AbstractCell> to state
			}
		}, Starbound.EXECUTOR)
	}

	override fun loadEntities(pos: ChunkPos): CompletableFuture<KOptional<Collection<AbstractEntity>>> {
		val chunkX = pos.x
		val chunkY = pos.y
		val key = ByteKey(2, (chunkX shr 8).toByte(), chunkX.toByte(), (chunkY shr 8).toByte(), chunkY.toByte())

		return load(key).thenApplyAsync(Function {
			it.map { readEntities(pos, it) }
		}, Starbound.EXECUTOR)
	}

	override fun loadMetadata(): CompletableFuture<KOptional<Metadata>> {
		return load(metadataKey).thenApplyAsync(Function {
			it.flatMap {
				val stream = DataInputStream(BufferedInputStream(InflaterInputStream(FastByteArrayInputStream(it))))

				val width = stream.readInt()
				val height = stream.readInt()

				val json = VersionedJson(stream)

				KOptional(Metadata(WorldGeometry(Vector2i(width, height), true, false), json))
			}
		}, Starbound.EXECUTOR)
	}

	override fun saveEntities(pos: ChunkPos, data: Collection<AbstractEntity>): Boolean {
		val chunkX = pos.x
		val chunkY = pos.y
		val key = ByteKey(2, (chunkX shr 8).toByte(), chunkX.toByte(), (chunkY shr 8).toByte(), chunkY.toByte())

		write(key, writeEntities(data))
		return true
	}

	override fun saveCells(pos: ChunkPos, data: Object2DArray<out AbstractCell>, state: ChunkState): Boolean {
		val buff = FastByteArrayOutputStream()
		val stream = DataOutputStream(BufferedOutputStream(DeflaterOutputStream(buff)))

		stream.writeVarInt(when (state) {
			ChunkState.FRESH -> 0
			ChunkState.EMPTY -> 0
			ChunkState.TERRAIN -> 1
			ChunkState.MICRO_DUNGEONS -> 2
			ChunkState.CAVE_LIQUID -> 3
			ChunkState.FULL -> 4
		})

		stream.writeVarInt(418)

		for (y in 0 until CHUNK_SIZE) {
			for (x in 0 until CHUNK_SIZE) {
				val cell = data.getOrNull(x, y) ?: AbstractCell.NULL
				cell.writeLegacy(stream)
			}
		}

		val chunkX = pos.x
		val chunkY = pos.y
		val key = ByteKey(1, (chunkX shr 8).toByte(), chunkX.toByte(), (chunkY shr 8).toByte(), chunkY.toByte())

		stream.close()
		write(key, buff.array.copyOf(buff.length))
		return true
	}

	override fun saveMetadata(data: Metadata): Boolean {
		val buff = FastByteArrayOutputStream()
		val stream = DataOutputStream(BufferedOutputStream(DeflaterOutputStream(buff)))

		stream.writeInt(data.geometry.size.x)
		stream.writeInt(data.geometry.size.y)
		data.data.write(stream)

		stream.close()
		write(metadataKey, buff.array.copyOf(buff.length))
		return true
	}

	class Memory(private val get: (ByteKey) -> ByteArray?, private val set: (ByteKey, ByteArray) -> Unit) : LegacyWorldStorage() {
		override fun load(at: ByteKey): CompletableFuture<KOptional<ByteArray>> {
			return CompletableFuture.completedFuture(KOptional.ofNullable(get(at)))
		}

		override fun write(at: ByteKey, value: ByteArray) {
			set(at, value)
		}

		override fun close() {}
		override fun sync() {}
	}

	class DB5(private val database: BTreeDB5) : LegacyWorldStorage() {
		private val carrier = CarriedExecutor(Starbound.IO_EXECUTOR)

		override fun load(at: ByteKey): CompletableFuture<KOptional<ByteArray>> {
			return CompletableFuture.supplyAsync(Supplier { database.read(at) }, carrier)
		}

		override fun write(at: ByteKey, value: ByteArray) {
			throw UnsupportedOperationException()
		}

		override fun sync() {
			// do nothing
		}

		override fun close() {
			carrier.execute { database.close() }
			carrier.wait(300L, TimeUnit.SECONDS)
		}
	}

	class SQL(path: File) : LegacyWorldStorage() {
		private val carrier = CarriedExecutor(Starbound.IO_EXECUTOR)
		private val connection = DriverManager.getConnection("jdbc:sqlite:${path.canonicalPath.replace('\\', '/')}")
		private val cleaner: Cleaner.Cleanable

		init {
			val connection = connection

			cleaner = Starbound.CLEANER.register(this) {
				/*connection.commit();*/
				connection.close()
			}

			connection.autoCommit = false

			connection.createStatement().use {
				it.execute("""CREATE TABLE IF NOT EXISTS `data` (
					|`key` BLOB NOT NULL PRIMARY KEY,
					|`value` BLOB NOT NULL
					|)""".trimMargin())
			}
		}

		override fun sync() {
			carrier.execute { connection.commit() }
		}

		// concurrent safety - load(), write() and close() are always called on one thread
		// and load()/write() will never be called after close()
		private val loader = connection.prepareStatement("SELECT `value` FROM `data` WHERE `key` = ? LIMIT 1")
		private val writer = connection.prepareStatement("REPLACE INTO `data` (`key`, `value`) VALUES (?, ?)")

		override fun load(at: ByteKey): CompletableFuture<KOptional<ByteArray>> {
			return CompletableFuture.supplyAsync(Supplier {
				loader.setBytes(1, at.toByteArray())

				loader.executeQuery().use {
					if (it.next()) {
						val blob = it.getBytes(1)
						KOptional(blob)
					} else {
						KOptional()
					}
				}
			}, carrier)
		}

		override fun write(at: ByteKey, value: ByteArray) {
			carrier.execute {
				writer.setBytes(1, at.toByteArray())
				writer.setBytes(2, value)
				writer.execute()
			}
		}

		override fun close() {
			carrier.execute { connection.commit() }
			carrier.execute { cleaner.clean() }
			carrier.wait(300L, TimeUnit.SECONDS)
		}
	}

	companion object {
		private val LOGGER = LogManager.getLogger()
		private val metadataKey = ByteKey(0, 0, 0, 0, 0)
	}
}
